#coding: utf8
import plask


class ActiveCond2D(plask.Solver):
    """
    ActiveCond2D(name="")
    
    Finite element thermal solver for 2D Cartesian geometry.
    """
    algorithm = None
    """
    Chosen matrix factorization algorithm
    """

    def compute(self, loops=0):
        """
        compute(self, loops=0)
        
            Run electrical calculations
        """

    cond = None
    """
    Junction conductivity function [S/m].
    
    This function should take current density [kA/cm²] and temperature [K]
    as arguments and return a conductivity [S/m]. In case,there is more than
    one junction you may set such function for any of them by using ``cond#``
    property, where # is the junction number (specified by a role ``junction#``
     or ``active#``).
    
    ``cond`` is an alias for ``cond0``.
    """

    convergence = None
    """
    Convergence method.
    
    If stable, covergence is slown down to ensure stability.
    """

    err = None
    """
    Maximum estimated error
    """

    geometry = None
    """
    Geometry provided to the solver
    """

    def get_capacitance(self):
        """
        get_capacitance(self)
        
            Get the structure capacitance.
            
            Return:
                Total capacitance [pF].
            
            Note:
                This method can only be used it there are exactly two boundary conditions
                specifying the voltage. Otherwise use :meth:`get_electrostatic_energy` to
                obtain the stored energy :math:`W` and compute the capacitance as:
                :math:`C = 2 \, W / U^2`, where :math:`U` is the applied voltage.
        """

    def get_electrostatic_energy(self):
        """
        get_electrostatic_energy(self)
        
            Get the energy stored in the electrostatic field in the analyzed structure.
            
            Return:
                Total electrostatic energy [J].
        """

    def get_total_current(self, nact=0):
        """
        get_total_current(self, nact=0)
        
            Get total current flowing through active region [mA]
        """

    def get_total_heat(self):
        """
        get_total_heat(self)
        
            Get the total heat produced by the current flowing in the structure.
            
            Return:
                Total produced heat [mW].
        """

    id = None
    """
    Id of the solver object. (read only)
    
    Example:
        >>> mysolver.id
        mysolver:category.type
    """

    inTemperature = None
    """
    Receiver of the temperature required for computations [K].
    
    
    You will find usage details in the documentation of the receiver class
    :class:`~plask.flow.TemperatureReceiver2D`.
    
    Example:
       Connect the reveiver to a provider from some other solver:
    
       >>> solver.inTemperature = other_solver.outTemperature
    
    See also:
    
       Receciver class: :class:`plask.flow.TemperatureReceiver2D`
    
       Provider class: :class:`plask.flow.TemperatureProvider2D`
    
       Data filter: :class:`plask.filter.TemperatureFilter2D`
    """

    include_empty = None
    """
    Should empty regions (e.g. air) be included into computation domain?
    """

    def initialize(self):
        """
        initialize(self)
        
            Initialize solver.
            
            This method manually initialized the solver and sets :attr:`initialized` to
            *True*. Normally calling it is not necessary, as each solver automatically
            initializes itself when needed.
            
            Returns:
                bool: solver :attr:`initialized` state prior to this method call.
        """

    initialized = None
    """
    True if the solver has been initialized. (read only)
    
    Solvers usually get initialized at the beginning of the computations.
    You can clean the initialization state and free the memory by calling
    the :meth:`invalidate` method.
    """

    def invalidate(self):
        """
        invalidate(self)
        
            Set the solver back to uninitialized state.
            
            This method frees the memory allocated by the solver and sets
            :attr:`initialized` to *False*.
        """

    itererr = None
    """
    Allowed residual iteration for iterative method
    """

    iterlim = None
    """
    Maximum number of iterations for iterative method
    """

    def load_xpl(self, xpl, manager):
        """
        load_xpl(self, xpl, manager)
        
            Load configuration from XPL reader.
            
            This method should be overriden in custom Python solvers.
            
            Example:
              .. code-block:: python
            
                 def load_xpl(self, xpl, manager):
                     for tag in xpl:
                         if tag == 'config':
                             self.a = tag['a']
                             self.b = tag.get('b', 0)
                             if 'c' in tag:
                                 self.c = tag['c']
                         if tag == 'combined':
                             for subtag in tag:
                                 if subtag == 'withtext':
                                     self.data = subtag.attrs
                                     # Text must be read last
                                     self.text = subtag.text
                         elif tag == 'geometry':
                             self.geometry = tag.getitem(manager.geo, 'ref')
        """

    logfreq = None
    """
    Frequency of iteration progress reporting
    """

    maxerr = None
    """
    Limit for the potential updates
    """

    mesh = None
    """
    Mesh provided to the solver
    """

    ncond = None
    """
    Conductivity of the n-contact
    """

    outConductivity = None
    """
    Provider of the computed electrical conductivity [S/m].
    
    
    outConductivity(mesh, interpolation='default')
    
    :param mesh mesh: Target mesh to get the field at.
    :param str interpolation: Requested interpolation method.
    
    :return: Data with the electrical conductivity on the specified mesh **[S/m]**.
    
    Example:
       Connect the provider to a receiver in some other solver:
    
       >>> other_solver.inConductivity = solver.outConductivity
    
       Obtain the provided field:
    
       >>> solver.outConductivity(mesh)
       <plask.Data at 0x1234567>
    
    See also:
    
       Provider class: :class:`plask.flow.ConductivityProvider2D`
    
       Receciver class: :class:`plask.flow.ConductivityReceiver2D`
    """

    outCurrentDensity = None
    """
    Provider of the computed current density [kA/cm²].
    
    
    outCurrentDensity(mesh, interpolation='default')
    
    :param mesh mesh: Target mesh to get the field at.
    :param str interpolation: Requested interpolation method.
    
    :return: Data with the current density on the specified mesh **[kA/cm²]**.
    
    Example:
       Connect the provider to a receiver in some other solver:
    
       >>> other_solver.inCurrentDensity = solver.outCurrentDensity
    
       Obtain the provided field:
    
       >>> solver.outCurrentDensity(mesh)
       <plask.Data at 0x1234567>
    
    See also:
    
       Provider class: :class:`plask.flow.CurrentDensityProvider2D`
    
       Receciver class: :class:`plask.flow.CurrentDensityReceiver2D`
    """

    outHeat = None
    """
    Provider of the computed heat sources density [W/m³].
    
    
    outHeat(mesh, interpolation='default')
    
    :param mesh mesh: Target mesh to get the field at.
    :param str interpolation: Requested interpolation method.
    
    :return: Data with the heat sources density on the specified mesh **[W/m³]**.
    
    Example:
       Connect the provider to a receiver in some other solver:
    
       >>> other_solver.inHeat = solver.outHeat
    
       Obtain the provided field:
    
       >>> solver.outHeat(mesh)
       <plask.Data at 0x1234567>
    
    See also:
    
       Provider class: :class:`plask.flow.HeatProvider2D`
    
       Receciver class: :class:`plask.flow.HeatReceiver2D`
    """

    outPotential = None
    """
    Not available in this solver. Use :attr:`outVoltage` instead.
    """

    outVoltage = None
    """
    Provider of the computed voltage [V].
    
    
    outVoltage(mesh, interpolation='default')
    
    :param mesh mesh: Target mesh to get the field at.
    :param str interpolation: Requested interpolation method.
    
    :return: Data with the voltage on the specified mesh **[V]**.
    
    Example:
       Connect the provider to a receiver in some other solver:
    
       >>> other_solver.inVoltage = solver.outVoltage
    
       Obtain the provided field:
    
       >>> solver.outVoltage(mesh)
       <plask.Data at 0x1234567>
    
    See also:
    
       Provider class: :class:`plask.flow.VoltageProvider2D`
    
       Receciver class: :class:`plask.flow.VoltageReceiver2D`
    """

    pcond = None
    """
    Conductivity of the p-contact
    """

    pnjcond = None
    """
    Default effective conductivity of the active region.
    
    Effective junction conductivity will be computed starting from this value.
    Note that the actual junction conductivity after convergence can be obtained
    with :attr:`outConductivity`.
    """

    start_cond = None
    """
    Default effective conductivity of the active region.
    
    Effective junction conductivity will be computed starting from this value.
    Note that the actual junction conductivity after convergence can be obtained
    with :attr:`outConductivity`.
    """

    voltage_boundary = None
    """
    Boundary conditions of the first kind (constant potential) 
    
    This field holds a list of boundary conditions for the solver. You may access
    and alter is elements a normal Python list. Each element is a special class
    that has two attributes:
    
    ============= ==================================================================
    :attr:`place` Boundary condition location (:class:`plask.mesh.RectangularBase2D.Boundary`).
    :attr:`value` Boundary condition value.
    ============= ==================================================================
    
    When you add new boundary condition, you may use two-argument ``append``, or
    ``prepend`` methods, or three-argument ``insert`` method, where you separately
    specify the place and the value. See the below example for clarification.
    
    Example:
        >>> solver.voltage_boundary.clear()
        >>> solver.voltage_boundary.append(solver.mesh.Bottom(), some_value)
        >>> solver.voltage_boundary[0].value = different_value
        >>> solver.voltage_boundary.insert(0, solver.mesh.Top(), new_value)
        >>> solver.voltage_boundary[1].value == different_value
        True
    """



class ActiveCond3D(plask.Solver):
    """
    ActiveCond3D(name="")
    
    Finite element thermal solver for 3D Cartesian geometry.
    """
    algorithm = None
    """
    Chosen matrix factorization algorithm
    """

    def compute(self, loops=0):
        """
        compute(self, loops=0)
        
            Run electrical calculations
        """

    cond = None
    """
    Junction conductivity function [S/m].
    
    This function should take current density [kA/cm²] and temperature [K]
    as arguments and return a conductivity [S/m]. In case,there is more than
    one junction you may set such function for any of them by using ``cond#``
    property, where # is the junction number (specified by a role ``junction#``
     or ``active#``).
    
    ``cond`` is an alias for ``cond0``.
    """

    convergence = None
    """
    Convergence method.
    
    If stable, covergence is slown down to ensure stability.
    """

    err = None
    """
    Maximum estimated error
    """

    geometry = None
    """
    Geometry provided to the solver
    """

    def get_capacitance(self):
        """
        get_capacitance(self)
        
            Get the structure capacitance.
            
            Return:
                Total capacitance [pF].
            
            Note:
                This method can only be used it there are exactly two boundary conditions
                specifying the voltage. Otherwise use :meth:`get_electrostatic_energy` to
                obtain the stored energy :math:`W` and compute the capacitance as:
                :math:`C = 2 \, W / U^2`, where :math:`U` is the applied voltage.
        """

    def get_electrostatic_energy(self):
        """
        get_electrostatic_energy(self)
        
            Get the energy stored in the electrostatic field in the analyzed structure.
            
            Return:
                Total electrostatic energy [J].
        """

    def get_total_current(self, nact=0):
        """
        get_total_current(self, nact=0)
        
            Get total current flowing through active region [mA]
        """

    def get_total_heat(self):
        """
        get_total_heat(self)
        
            Get the total heat produced by the current flowing in the structure.
            
            Return:
                Total produced heat [mW].
        """

    id = None
    """
    Id of the solver object. (read only)
    
    Example:
        >>> mysolver.id
        mysolver:category.type
    """

    inTemperature = None
    """
    Receiver of the temperature required for computations [K].
    
    
    You will find usage details in the documentation of the receiver class
    :class:`~plask.flow.TemperatureReceiver3D`.
    
    Example:
       Connect the reveiver to a provider from some other solver:
    
       >>> solver.inTemperature = other_solver.outTemperature
    
    See also:
    
       Receciver class: :class:`plask.flow.TemperatureReceiver3D`
    
       Provider class: :class:`plask.flow.TemperatureProvider3D`
    
       Data filter: :class:`plask.filter.TemperatureFilter3D`
    """

    include_empty = None
    """
    Should empty regions (e.g. air) be included into computation domain?
    """

    def initialize(self):
        """
        initialize(self)
        
            Initialize solver.
            
            This method manually initialized the solver and sets :attr:`initialized` to
            *True*. Normally calling it is not necessary, as each solver automatically
            initializes itself when needed.
            
            Returns:
                bool: solver :attr:`initialized` state prior to this method call.
        """

    initialized = None
    """
    True if the solver has been initialized. (read only)
    
    Solvers usually get initialized at the beginning of the computations.
    You can clean the initialization state and free the memory by calling
    the :meth:`invalidate` method.
    """

    def invalidate(self):
        """
        invalidate(self)
        
            Set the solver back to uninitialized state.
            
            This method frees the memory allocated by the solver and sets
            :attr:`initialized` to *False*.
        """

    itererr = None
    """
    Allowed residual iteration for iterative method
    """

    iterlim = None
    """
    Maximum number of iterations for iterative method
    """

    def load_xpl(self, xpl, manager):
        """
        load_xpl(self, xpl, manager)
        
            Load configuration from XPL reader.
            
            This method should be overriden in custom Python solvers.
            
            Example:
              .. code-block:: python
            
                 def load_xpl(self, xpl, manager):
                     for tag in xpl:
                         if tag == 'config':
                             self.a = tag['a']
                             self.b = tag.get('b', 0)
                             if 'c' in tag:
                                 self.c = tag['c']
                         if tag == 'combined':
                             for subtag in tag:
                                 if subtag == 'withtext':
                                     self.data = subtag.attrs
                                     # Text must be read last
                                     self.text = subtag.text
                         elif tag == 'geometry':
                             self.geometry = tag.getitem(manager.geo, 'ref')
        """

    logfreq = None
    """
    Frequency of iteration progress reporting
    """

    maxerr = None
    """
    Limit for the potential updates
    """

    mesh = None
    """
    Mesh provided to the solver
    """

    ncond = None
    """
    Conductivity of the n-contact
    """

    outConductivity = None
    """
    Provider of the computed electrical conductivity [S/m].
    
    
    outConductivity(mesh, interpolation='default')
    
    :param mesh mesh: Target mesh to get the field at.
    :param str interpolation: Requested interpolation method.
    
    :return: Data with the electrical conductivity on the specified mesh **[S/m]**.
    
    Example:
       Connect the provider to a receiver in some other solver:
    
       >>> other_solver.inConductivity = solver.outConductivity
    
       Obtain the provided field:
    
       >>> solver.outConductivity(mesh)
       <plask.Data at 0x1234567>
    
    See also:
    
       Provider class: :class:`plask.flow.ConductivityProvider3D`
    
       Receciver class: :class:`plask.flow.ConductivityReceiver3D`
    """

    outCurrentDensity = None
    """
    Provider of the computed current density [kA/cm²].
    
    
    outCurrentDensity(mesh, interpolation='default')
    
    :param mesh mesh: Target mesh to get the field at.
    :param str interpolation: Requested interpolation method.
    
    :return: Data with the current density on the specified mesh **[kA/cm²]**.
    
    Example:
       Connect the provider to a receiver in some other solver:
    
       >>> other_solver.inCurrentDensity = solver.outCurrentDensity
    
       Obtain the provided field:
    
       >>> solver.outCurrentDensity(mesh)
       <plask.Data at 0x1234567>
    
    See also:
    
       Provider class: :class:`plask.flow.CurrentDensityProvider3D`
    
       Receciver class: :class:`plask.flow.CurrentDensityReceiver3D`
    """

    outHeat = None
    """
    Provider of the computed heat sources density [W/m³].
    
    
    outHeat(mesh, interpolation='default')
    
    :param mesh mesh: Target mesh to get the field at.
    :param str interpolation: Requested interpolation method.
    
    :return: Data with the heat sources density on the specified mesh **[W/m³]**.
    
    Example:
       Connect the provider to a receiver in some other solver:
    
       >>> other_solver.inHeat = solver.outHeat
    
       Obtain the provided field:
    
       >>> solver.outHeat(mesh)
       <plask.Data at 0x1234567>
    
    See also:
    
       Provider class: :class:`plask.flow.HeatProvider3D`
    
       Receciver class: :class:`plask.flow.HeatReceiver3D`
    """

    outPotential = None
    """
    Not available in this solver. Use :attr:`outVoltage` instead.
    """

    outVoltage = None
    """
    Provider of the computed voltage [V].
    
    
    outVoltage(mesh, interpolation='default')
    
    :param mesh mesh: Target mesh to get the field at.
    :param str interpolation: Requested interpolation method.
    
    :return: Data with the voltage on the specified mesh **[V]**.
    
    Example:
       Connect the provider to a receiver in some other solver:
    
       >>> other_solver.inVoltage = solver.outVoltage
    
       Obtain the provided field:
    
       >>> solver.outVoltage(mesh)
       <plask.Data at 0x1234567>
    
    See also:
    
       Provider class: :class:`plask.flow.VoltageProvider3D`
    
       Receciver class: :class:`plask.flow.VoltageReceiver3D`
    """

    pcond = None
    """
    Conductivity of the p-contact
    """

    pnjcond = None
    """
    Default effective conductivity of the active region.
    
    Effective junction conductivity will be computed starting from this value.
    Note that the actual junction conductivity after convergence can be obtained
    with :attr:`outConductivity`.
    """

    start_cond = None
    """
    Default effective conductivity of the active region.
    
    Effective junction conductivity will be computed starting from this value.
    Note that the actual junction conductivity after convergence can be obtained
    with :attr:`outConductivity`.
    """

    voltage_boundary = None
    """
    Boundary conditions of the first kind (constant potential) 
    
    This field holds a list of boundary conditions for the solver. You may access
    and alter is elements a normal Python list. Each element is a special class
    that has two attributes:
    
    ============= ==================================================================
    :attr:`place` Boundary condition location (:class:`plask.mesh.RectangularBase3D.Boundary`).
    :attr:`value` Boundary condition value.
    ============= ==================================================================
    
    When you add new boundary condition, you may use two-argument ``append``, or
    ``prepend`` methods, or three-argument ``insert`` method, where you separately
    specify the place and the value. See the below example for clarification.
    
    Example:
        >>> solver.voltage_boundary.clear()
        >>> solver.voltage_boundary.append(solver.mesh.Bottom(), some_value)
        >>> solver.voltage_boundary[0].value = different_value
        >>> solver.voltage_boundary.insert(0, solver.mesh.Top(), new_value)
        >>> solver.voltage_boundary[1].value == different_value
        True
    """



class ActiveCondCyl(plask.Solver):
    """
    ActiveCondCyl(name="")
    
    Finite element thermal solver for 2D cylindrical geometry.
    """
    algorithm = None
    """
    Chosen matrix factorization algorithm
    """

    def compute(self, loops=0):
        """
        compute(self, loops=0)
        
            Run electrical calculations
        """

    cond = None
    """
    Junction conductivity function [S/m].
    
    This function should take current density [kA/cm²] and temperature [K]
    as arguments and return a conductivity [S/m]. In case,there is more than
    one junction you may set such function for any of them by using ``cond#``
    property, where # is the junction number (specified by a role ``junction#``
     or ``active#``).
    
    ``cond`` is an alias for ``cond0``.
    """

    convergence = None
    """
    Convergence method.
    
    If stable, covergence is slown down to ensure stability.
    """

    err = None
    """
    Maximum estimated error
    """

    geometry = None
    """
    Geometry provided to the solver
    """

    def get_capacitance(self):
        """
        get_capacitance(self)
        
            Get the structure capacitance.
            
            Return:
                Total capacitance [pF].
            
            Note:
                This method can only be used it there are exactly two boundary conditions
                specifying the voltage. Otherwise use :meth:`get_electrostatic_energy` to
                obtain the stored energy :math:`W` and compute the capacitance as:
                :math:`C = 2 \, W / U^2`, where :math:`U` is the applied voltage.
        """

    def get_electrostatic_energy(self):
        """
        get_electrostatic_energy(self)
        
            Get the energy stored in the electrostatic field in the analyzed structure.
            
            Return:
                Total electrostatic energy [J].
        """

    def get_total_current(self, nact=0):
        """
        get_total_current(self, nact=0)
        
            Get total current flowing through active region [mA]
        """

    def get_total_heat(self):
        """
        get_total_heat(self)
        
            Get the total heat produced by the current flowing in the structure.
            
            Return:
                Total produced heat [mW].
        """

    id = None
    """
    Id of the solver object. (read only)
    
    Example:
        >>> mysolver.id
        mysolver:category.type
    """

    inTemperature = None
    """
    Receiver of the temperature required for computations [K].
    
    
    You will find usage details in the documentation of the receiver class
    :class:`~plask.flow.TemperatureReceiverCyl`.
    
    Example:
       Connect the reveiver to a provider from some other solver:
    
       >>> solver.inTemperature = other_solver.outTemperature
    
    See also:
    
       Receciver class: :class:`plask.flow.TemperatureReceiverCyl`
    
       Provider class: :class:`plask.flow.TemperatureProviderCyl`
    
       Data filter: :class:`plask.filter.TemperatureFilterCyl`
    """

    include_empty = None
    """
    Should empty regions (e.g. air) be included into computation domain?
    """

    def initialize(self):
        """
        initialize(self)
        
            Initialize solver.
            
            This method manually initialized the solver and sets :attr:`initialized` to
            *True*. Normally calling it is not necessary, as each solver automatically
            initializes itself when needed.
            
            Returns:
                bool: solver :attr:`initialized` state prior to this method call.
        """

    initialized = None
    """
    True if the solver has been initialized. (read only)
    
    Solvers usually get initialized at the beginning of the computations.
    You can clean the initialization state and free the memory by calling
    the :meth:`invalidate` method.
    """

    def invalidate(self):
        """
        invalidate(self)
        
            Set the solver back to uninitialized state.
            
            This method frees the memory allocated by the solver and sets
            :attr:`initialized` to *False*.
        """

    itererr = None
    """
    Allowed residual iteration for iterative method
    """

    iterlim = None
    """
    Maximum number of iterations for iterative method
    """

    def load_xpl(self, xpl, manager):
        """
        load_xpl(self, xpl, manager)
        
            Load configuration from XPL reader.
            
            This method should be overriden in custom Python solvers.
            
            Example:
              .. code-block:: python
            
                 def load_xpl(self, xpl, manager):
                     for tag in xpl:
                         if tag == 'config':
                             self.a = tag['a']
                             self.b = tag.get('b', 0)
                             if 'c' in tag:
                                 self.c = tag['c']
                         if tag == 'combined':
                             for subtag in tag:
                                 if subtag == 'withtext':
                                     self.data = subtag.attrs
                                     # Text must be read last
                                     self.text = subtag.text
                         elif tag == 'geometry':
                             self.geometry = tag.getitem(manager.geo, 'ref')
        """

    logfreq = None
    """
    Frequency of iteration progress reporting
    """

    maxerr = None
    """
    Limit for the potential updates
    """

    mesh = None
    """
    Mesh provided to the solver
    """

    ncond = None
    """
    Conductivity of the n-contact
    """

    outConductivity = None
    """
    Provider of the computed electrical conductivity [S/m].
    
    
    outConductivity(mesh, interpolation='default')
    
    :param mesh mesh: Target mesh to get the field at.
    :param str interpolation: Requested interpolation method.
    
    :return: Data with the electrical conductivity on the specified mesh **[S/m]**.
    
    Example:
       Connect the provider to a receiver in some other solver:
    
       >>> other_solver.inConductivity = solver.outConductivity
    
       Obtain the provided field:
    
       >>> solver.outConductivity(mesh)
       <plask.Data at 0x1234567>
    
    See also:
    
       Provider class: :class:`plask.flow.ConductivityProviderCyl`
    
       Receciver class: :class:`plask.flow.ConductivityReceiverCyl`
    """

    outCurrentDensity = None
    """
    Provider of the computed current density [kA/cm²].
    
    
    outCurrentDensity(mesh, interpolation='default')
    
    :param mesh mesh: Target mesh to get the field at.
    :param str interpolation: Requested interpolation method.
    
    :return: Data with the current density on the specified mesh **[kA/cm²]**.
    
    Example:
       Connect the provider to a receiver in some other solver:
    
       >>> other_solver.inCurrentDensity = solver.outCurrentDensity
    
       Obtain the provided field:
    
       >>> solver.outCurrentDensity(mesh)
       <plask.Data at 0x1234567>
    
    See also:
    
       Provider class: :class:`plask.flow.CurrentDensityProviderCyl`
    
       Receciver class: :class:`plask.flow.CurrentDensityReceiverCyl`
    """

    outHeat = None
    """
    Provider of the computed heat sources density [W/m³].
    
    
    outHeat(mesh, interpolation='default')
    
    :param mesh mesh: Target mesh to get the field at.
    :param str interpolation: Requested interpolation method.
    
    :return: Data with the heat sources density on the specified mesh **[W/m³]**.
    
    Example:
       Connect the provider to a receiver in some other solver:
    
       >>> other_solver.inHeat = solver.outHeat
    
       Obtain the provided field:
    
       >>> solver.outHeat(mesh)
       <plask.Data at 0x1234567>
    
    See also:
    
       Provider class: :class:`plask.flow.HeatProviderCyl`
    
       Receciver class: :class:`plask.flow.HeatReceiverCyl`
    """

    outPotential = None
    """
    Not available in this solver. Use :attr:`outVoltage` instead.
    """

    outVoltage = None
    """
    Provider of the computed voltage [V].
    
    
    outVoltage(mesh, interpolation='default')
    
    :param mesh mesh: Target mesh to get the field at.
    :param str interpolation: Requested interpolation method.
    
    :return: Data with the voltage on the specified mesh **[V]**.
    
    Example:
       Connect the provider to a receiver in some other solver:
    
       >>> other_solver.inVoltage = solver.outVoltage
    
       Obtain the provided field:
    
       >>> solver.outVoltage(mesh)
       <plask.Data at 0x1234567>
    
    See also:
    
       Provider class: :class:`plask.flow.VoltageProviderCyl`
    
       Receciver class: :class:`plask.flow.VoltageReceiverCyl`
    """

    pcond = None
    """
    Conductivity of the p-contact
    """

    pnjcond = None
    """
    Default effective conductivity of the active region.
    
    Effective junction conductivity will be computed starting from this value.
    Note that the actual junction conductivity after convergence can be obtained
    with :attr:`outConductivity`.
    """

    start_cond = None
    """
    Default effective conductivity of the active region.
    
    Effective junction conductivity will be computed starting from this value.
    Note that the actual junction conductivity after convergence can be obtained
    with :attr:`outConductivity`.
    """

    voltage_boundary = None
    """
    Boundary conditions of the first kind (constant potential) 
    
    This field holds a list of boundary conditions for the solver. You may access
    and alter is elements a normal Python list. Each element is a special class
    that has two attributes:
    
    ============= ==================================================================
    :attr:`place` Boundary condition location (:class:`plask.mesh.RectangularBase2D.Boundary`).
    :attr:`value` Boundary condition value.
    ============= ==================================================================
    
    When you add new boundary condition, you may use two-argument ``append``, or
    ``prepend`` methods, or three-argument ``insert`` method, where you separately
    specify the place and the value. See the below example for clarification.
    
    Example:
        >>> solver.voltage_boundary.clear()
        >>> solver.voltage_boundary.append(solver.mesh.Bottom(), some_value)
        >>> solver.voltage_boundary[0].value = different_value
        >>> solver.voltage_boundary.insert(0, solver.mesh.Top(), new_value)
        >>> solver.voltage_boundary[1].value == different_value
        True
    """



class Shockley2D(plask.Solver):
    """
    Shockley2D(name="")
    
    Finite element thermal solver for 2D Cartesian geometry.
    """
    algorithm = None
    """
    Chosen matrix factorization algorithm
    """

    beta = None
    """
    Junction coefficient [1/V].
    
    In case, there is more than one junction you may set $\beta$ parameter for any
    of them by using ``beta#`` property, where # is the junction number (specified
    by a role ``junction#`` or ``active#``).
    
    ``beta`` is an alias for ``beta0``.
    """

    def compute(self, loops=0):
        """
        compute(self, loops=0)
        
            Run electrical calculations
        """

    convergence = None
    """
    Convergence method.
    
    If stable, covergence is slown down to ensure stability.
    """

    err = None
    """
    Maximum estimated error
    """

    geometry = None
    """
    Geometry provided to the solver
    """

    def get_capacitance(self):
        """
        get_capacitance(self)
        
            Get the structure capacitance.
            
            Return:
                Total capacitance [pF].
            
            Note:
                This method can only be used it there are exactly two boundary conditions
                specifying the voltage. Otherwise use :meth:`get_electrostatic_energy` to
                obtain the stored energy :math:`W` and compute the capacitance as:
                :math:`C = 2 \, W / U^2`, where :math:`U` is the applied voltage.
        """

    def get_electrostatic_energy(self):
        """
        get_electrostatic_energy(self)
        
            Get the energy stored in the electrostatic field in the analyzed structure.
            
            Return:
                Total electrostatic energy [J].
        """

    def get_total_current(self, nact=0):
        """
        get_total_current(self, nact=0)
        
            Get total current flowing through active region [mA]
        """

    def get_total_heat(self):
        """
        get_total_heat(self)
        
            Get the total heat produced by the current flowing in the structure.
            
            Return:
                Total produced heat [mW].
        """

    id = None
    """
    Id of the solver object. (read only)
    
    Example:
        >>> mysolver.id
        mysolver:category.type
    """

    inTemperature = None
    """
    Receiver of the temperature required for computations [K].
    
    
    You will find usage details in the documentation of the receiver class
    :class:`~plask.flow.TemperatureReceiver2D`.
    
    Example:
       Connect the reveiver to a provider from some other solver:
    
       >>> solver.inTemperature = other_solver.outTemperature
    
    See also:
    
       Receciver class: :class:`plask.flow.TemperatureReceiver2D`
    
       Provider class: :class:`plask.flow.TemperatureProvider2D`
    
       Data filter: :class:`plask.filter.TemperatureFilter2D`
    """

    include_empty = None
    """
    Should empty regions (e.g. air) be included into computation domain?
    """

    def initialize(self):
        """
        initialize(self)
        
            Initialize solver.
            
            This method manually initialized the solver and sets :attr:`initialized` to
            *True*. Normally calling it is not necessary, as each solver automatically
            initializes itself when needed.
            
            Returns:
                bool: solver :attr:`initialized` state prior to this method call.
        """

    initialized = None
    """
    True if the solver has been initialized. (read only)
    
    Solvers usually get initialized at the beginning of the computations.
    You can clean the initialization state and free the memory by calling
    the :meth:`invalidate` method.
    """

    def invalidate(self):
        """
        invalidate(self)
        
            Set the solver back to uninitialized state.
            
            This method frees the memory allocated by the solver and sets
            :attr:`initialized` to *False*.
        """

    itererr = None
    """
    Allowed residual iteration for iterative method
    """

    iterlim = None
    """
    Maximum number of iterations for iterative method
    """

    js = None
    """
    Reverse bias current density [A/m\ :sup:`2`\ ].
    
    In case, there is more than one junction you may set $j_s$ parameter for any
    of them by using ``js#`` property, where # is the junction number (specified
    by a role ``junction#`` or ``active#``).
    
    ``js`` is an alias for ``js0``.
    """

    def load_xpl(self, xpl, manager):
        """
        load_xpl(self, xpl, manager)
        
            Load configuration from XPL reader.
            
            This method should be overriden in custom Python solvers.
            
            Example:
              .. code-block:: python
            
                 def load_xpl(self, xpl, manager):
                     for tag in xpl:
                         if tag == 'config':
                             self.a = tag['a']
                             self.b = tag.get('b', 0)
                             if 'c' in tag:
                                 self.c = tag['c']
                         if tag == 'combined':
                             for subtag in tag:
                                 if subtag == 'withtext':
                                     self.data = subtag.attrs
                                     # Text must be read last
                                     self.text = subtag.text
                         elif tag == 'geometry':
                             self.geometry = tag.getitem(manager.geo, 'ref')
        """

    logfreq = None
    """
    Frequency of iteration progress reporting
    """

    maxerr = None
    """
    Limit for the potential updates
    """

    mesh = None
    """
    Mesh provided to the solver
    """

    ncond = None
    """
    Conductivity of the n-contact
    """

    outConductivity = None
    """
    Provider of the computed electrical conductivity [S/m].
    
    
    outConductivity(mesh, interpolation='default')
    
    :param mesh mesh: Target mesh to get the field at.
    :param str interpolation: Requested interpolation method.
    
    :return: Data with the electrical conductivity on the specified mesh **[S/m]**.
    
    Example:
       Connect the provider to a receiver in some other solver:
    
       >>> other_solver.inConductivity = solver.outConductivity
    
       Obtain the provided field:
    
       >>> solver.outConductivity(mesh)
       <plask.Data at 0x1234567>
    
    See also:
    
       Provider class: :class:`plask.flow.ConductivityProvider2D`
    
       Receciver class: :class:`plask.flow.ConductivityReceiver2D`
    """

    outCurrentDensity = None
    """
    Provider of the computed current density [kA/cm²].
    
    
    outCurrentDensity(mesh, interpolation='default')
    
    :param mesh mesh: Target mesh to get the field at.
    :param str interpolation: Requested interpolation method.
    
    :return: Data with the current density on the specified mesh **[kA/cm²]**.
    
    Example:
       Connect the provider to a receiver in some other solver:
    
       >>> other_solver.inCurrentDensity = solver.outCurrentDensity
    
       Obtain the provided field:
    
       >>> solver.outCurrentDensity(mesh)
       <plask.Data at 0x1234567>
    
    See also:
    
       Provider class: :class:`plask.flow.CurrentDensityProvider2D`
    
       Receciver class: :class:`plask.flow.CurrentDensityReceiver2D`
    """

    outHeat = None
    """
    Provider of the computed heat sources density [W/m³].
    
    
    outHeat(mesh, interpolation='default')
    
    :param mesh mesh: Target mesh to get the field at.
    :param str interpolation: Requested interpolation method.
    
    :return: Data with the heat sources density on the specified mesh **[W/m³]**.
    
    Example:
       Connect the provider to a receiver in some other solver:
    
       >>> other_solver.inHeat = solver.outHeat
    
       Obtain the provided field:
    
       >>> solver.outHeat(mesh)
       <plask.Data at 0x1234567>
    
    See also:
    
       Provider class: :class:`plask.flow.HeatProvider2D`
    
       Receciver class: :class:`plask.flow.HeatReceiver2D`
    """

    outPotential = None
    """
    Not available in this solver. Use :attr:`outVoltage` instead.
    """

    outVoltage = None
    """
    Provider of the computed voltage [V].
    
    
    outVoltage(mesh, interpolation='default')
    
    :param mesh mesh: Target mesh to get the field at.
    :param str interpolation: Requested interpolation method.
    
    :return: Data with the voltage on the specified mesh **[V]**.
    
    Example:
       Connect the provider to a receiver in some other solver:
    
       >>> other_solver.inVoltage = solver.outVoltage
    
       Obtain the provided field:
    
       >>> solver.outVoltage(mesh)
       <plask.Data at 0x1234567>
    
    See also:
    
       Provider class: :class:`plask.flow.VoltageProvider2D`
    
       Receciver class: :class:`plask.flow.VoltageReceiver2D`
    """

    pcond = None
    """
    Conductivity of the p-contact
    """

    pnjcond = None
    """
    Default effective conductivity of the active region.
    
    Effective junction conductivity will be computed starting from this value.
    Note that the actual junction conductivity after convergence can be obtained
    with :attr:`outConductivity`.
    """

    start_cond = None
    """
    Default effective conductivity of the active region.
    
    Effective junction conductivity will be computed starting from this value.
    Note that the actual junction conductivity after convergence can be obtained
    with :attr:`outConductivity`.
    """

    voltage_boundary = None
    """
    Boundary conditions of the first kind (constant potential) 
    
    This field holds a list of boundary conditions for the solver. You may access
    and alter is elements a normal Python list. Each element is a special class
    that has two attributes:
    
    ============= ==================================================================
    :attr:`place` Boundary condition location (:class:`plask.mesh.RectangularBase2D.Boundary`).
    :attr:`value` Boundary condition value.
    ============= ==================================================================
    
    When you add new boundary condition, you may use two-argument ``append``, or
    ``prepend`` methods, or three-argument ``insert`` method, where you separately
    specify the place and the value. See the below example for clarification.
    
    Example:
        >>> solver.voltage_boundary.clear()
        >>> solver.voltage_boundary.append(solver.mesh.Bottom(), some_value)
        >>> solver.voltage_boundary[0].value = different_value
        >>> solver.voltage_boundary.insert(0, solver.mesh.Top(), new_value)
        >>> solver.voltage_boundary[1].value == different_value
        True
    """



class Shockley3D(plask.Solver):
    """
    Shockley3D(name="")
    
    Finite element thermal solver for 3D Cartesian geometry.
    """
    algorithm = None
    """
    Chosen matrix factorization algorithm
    """

    beta = None
    """
    Junction coefficient [1/V].
    
    In case, there is more than one junction you may set $\beta$ parameter for any
    of them by using ``beta#`` property, where # is the junction number (specified
    by a role ``junction#`` or ``active#``).
    
    ``beta`` is an alias for ``beta0``.
    """

    def compute(self, loops=0):
        """
        compute(self, loops=0)
        
            Run electrical calculations
        """

    convergence = None
    """
    Convergence method.
    
    If stable, covergence is slown down to ensure stability.
    """

    err = None
    """
    Maximum estimated error
    """

    geometry = None
    """
    Geometry provided to the solver
    """

    def get_capacitance(self):
        """
        get_capacitance(self)
        
            Get the structure capacitance.
            
            Return:
                Total capacitance [pF].
            
            Note:
                This method can only be used it there are exactly two boundary conditions
                specifying the voltage. Otherwise use :meth:`get_electrostatic_energy` to
                obtain the stored energy :math:`W` and compute the capacitance as:
                :math:`C = 2 \, W / U^2`, where :math:`U` is the applied voltage.
        """

    def get_electrostatic_energy(self):
        """
        get_electrostatic_energy(self)
        
            Get the energy stored in the electrostatic field in the analyzed structure.
            
            Return:
                Total electrostatic energy [J].
        """

    def get_total_current(self, nact=0):
        """
        get_total_current(self, nact=0)
        
            Get total current flowing through active region [mA]
        """

    def get_total_heat(self):
        """
        get_total_heat(self)
        
            Get the total heat produced by the current flowing in the structure.
            
            Return:
                Total produced heat [mW].
        """

    id = None
    """
    Id of the solver object. (read only)
    
    Example:
        >>> mysolver.id
        mysolver:category.type
    """

    inTemperature = None
    """
    Receiver of the temperature required for computations [K].
    
    
    You will find usage details in the documentation of the receiver class
    :class:`~plask.flow.TemperatureReceiver3D`.
    
    Example:
       Connect the reveiver to a provider from some other solver:
    
       >>> solver.inTemperature = other_solver.outTemperature
    
    See also:
    
       Receciver class: :class:`plask.flow.TemperatureReceiver3D`
    
       Provider class: :class:`plask.flow.TemperatureProvider3D`
    
       Data filter: :class:`plask.filter.TemperatureFilter3D`
    """

    include_empty = None
    """
    Should empty regions (e.g. air) be included into computation domain?
    """

    def initialize(self):
        """
        initialize(self)
        
            Initialize solver.
            
            This method manually initialized the solver and sets :attr:`initialized` to
            *True*. Normally calling it is not necessary, as each solver automatically
            initializes itself when needed.
            
            Returns:
                bool: solver :attr:`initialized` state prior to this method call.
        """

    initialized = None
    """
    True if the solver has been initialized. (read only)
    
    Solvers usually get initialized at the beginning of the computations.
    You can clean the initialization state and free the memory by calling
    the :meth:`invalidate` method.
    """

    def invalidate(self):
        """
        invalidate(self)
        
            Set the solver back to uninitialized state.
            
            This method frees the memory allocated by the solver and sets
            :attr:`initialized` to *False*.
        """

    itererr = None
    """
    Allowed residual iteration for iterative method
    """

    iterlim = None
    """
    Maximum number of iterations for iterative method
    """

    js = None
    """
    Reverse bias current density [A/m\ :sup:`2`\ ].
    
    In case, there is more than one junction you may set $j_s$ parameter for any
    of them by using ``js#`` property, where # is the junction number (specified
    by a role ``junction#`` or ``active#``).
    
    ``js`` is an alias for ``js0``.
    """

    def load_xpl(self, xpl, manager):
        """
        load_xpl(self, xpl, manager)
        
            Load configuration from XPL reader.
            
            This method should be overriden in custom Python solvers.
            
            Example:
              .. code-block:: python
            
                 def load_xpl(self, xpl, manager):
                     for tag in xpl:
                         if tag == 'config':
                             self.a = tag['a']
                             self.b = tag.get('b', 0)
                             if 'c' in tag:
                                 self.c = tag['c']
                         if tag == 'combined':
                             for subtag in tag:
                                 if subtag == 'withtext':
                                     self.data = subtag.attrs
                                     # Text must be read last
                                     self.text = subtag.text
                         elif tag == 'geometry':
                             self.geometry = tag.getitem(manager.geo, 'ref')
        """

    logfreq = None
    """
    Frequency of iteration progress reporting
    """

    maxerr = None
    """
    Limit for the potential updates
    """

    mesh = None
    """
    Mesh provided to the solver
    """

    ncond = None
    """
    Conductivity of the n-contact
    """

    outConductivity = None
    """
    Provider of the computed electrical conductivity [S/m].
    
    
    outConductivity(mesh, interpolation='default')
    
    :param mesh mesh: Target mesh to get the field at.
    :param str interpolation: Requested interpolation method.
    
    :return: Data with the electrical conductivity on the specified mesh **[S/m]**.
    
    Example:
       Connect the provider to a receiver in some other solver:
    
       >>> other_solver.inConductivity = solver.outConductivity
    
       Obtain the provided field:
    
       >>> solver.outConductivity(mesh)
       <plask.Data at 0x1234567>
    
    See also:
    
       Provider class: :class:`plask.flow.ConductivityProvider3D`
    
       Receciver class: :class:`plask.flow.ConductivityReceiver3D`
    """

    outCurrentDensity = None
    """
    Provider of the computed current density [kA/cm²].
    
    
    outCurrentDensity(mesh, interpolation='default')
    
    :param mesh mesh: Target mesh to get the field at.
    :param str interpolation: Requested interpolation method.
    
    :return: Data with the current density on the specified mesh **[kA/cm²]**.
    
    Example:
       Connect the provider to a receiver in some other solver:
    
       >>> other_solver.inCurrentDensity = solver.outCurrentDensity
    
       Obtain the provided field:
    
       >>> solver.outCurrentDensity(mesh)
       <plask.Data at 0x1234567>
    
    See also:
    
       Provider class: :class:`plask.flow.CurrentDensityProvider3D`
    
       Receciver class: :class:`plask.flow.CurrentDensityReceiver3D`
    """

    outHeat = None
    """
    Provider of the computed heat sources density [W/m³].
    
    
    outHeat(mesh, interpolation='default')
    
    :param mesh mesh: Target mesh to get the field at.
    :param str interpolation: Requested interpolation method.
    
    :return: Data with the heat sources density on the specified mesh **[W/m³]**.
    
    Example:
       Connect the provider to a receiver in some other solver:
    
       >>> other_solver.inHeat = solver.outHeat
    
       Obtain the provided field:
    
       >>> solver.outHeat(mesh)
       <plask.Data at 0x1234567>
    
    See also:
    
       Provider class: :class:`plask.flow.HeatProvider3D`
    
       Receciver class: :class:`plask.flow.HeatReceiver3D`
    """

    outPotential = None
    """
    Not available in this solver. Use :attr:`outVoltage` instead.
    """

    outVoltage = None
    """
    Provider of the computed voltage [V].
    
    
    outVoltage(mesh, interpolation='default')
    
    :param mesh mesh: Target mesh to get the field at.
    :param str interpolation: Requested interpolation method.
    
    :return: Data with the voltage on the specified mesh **[V]**.
    
    Example:
       Connect the provider to a receiver in some other solver:
    
       >>> other_solver.inVoltage = solver.outVoltage
    
       Obtain the provided field:
    
       >>> solver.outVoltage(mesh)
       <plask.Data at 0x1234567>
    
    See also:
    
       Provider class: :class:`plask.flow.VoltageProvider3D`
    
       Receciver class: :class:`plask.flow.VoltageReceiver3D`
    """

    pcond = None
    """
    Conductivity of the p-contact
    """

    pnjcond = None
    """
    Default effective conductivity of the active region.
    
    Effective junction conductivity will be computed starting from this value.
    Note that the actual junction conductivity after convergence can be obtained
    with :attr:`outConductivity`.
    """

    start_cond = None
    """
    Default effective conductivity of the active region.
    
    Effective junction conductivity will be computed starting from this value.
    Note that the actual junction conductivity after convergence can be obtained
    with :attr:`outConductivity`.
    """

    voltage_boundary = None
    """
    Boundary conditions of the first kind (constant potential) 
    
    This field holds a list of boundary conditions for the solver. You may access
    and alter is elements a normal Python list. Each element is a special class
    that has two attributes:
    
    ============= ==================================================================
    :attr:`place` Boundary condition location (:class:`plask.mesh.RectangularBase3D.Boundary`).
    :attr:`value` Boundary condition value.
    ============= ==================================================================
    
    When you add new boundary condition, you may use two-argument ``append``, or
    ``prepend`` methods, or three-argument ``insert`` method, where you separately
    specify the place and the value. See the below example for clarification.
    
    Example:
        >>> solver.voltage_boundary.clear()
        >>> solver.voltage_boundary.append(solver.mesh.Bottom(), some_value)
        >>> solver.voltage_boundary[0].value = different_value
        >>> solver.voltage_boundary.insert(0, solver.mesh.Top(), new_value)
        >>> solver.voltage_boundary[1].value == different_value
        True
    """



class ShockleyCyl(plask.Solver):
    """
    ShockleyCyl(name="")
    
    Finite element thermal solver for 2D cylindrical geometry.
    """
    algorithm = None
    """
    Chosen matrix factorization algorithm
    """

    beta = None
    """
    Junction coefficient [1/V].
    
    In case, there is more than one junction you may set $\beta$ parameter for any
    of them by using ``beta#`` property, where # is the junction number (specified
    by a role ``junction#`` or ``active#``).
    
    ``beta`` is an alias for ``beta0``.
    """

    def compute(self, loops=0):
        """
        compute(self, loops=0)
        
            Run electrical calculations
        """

    convergence = None
    """
    Convergence method.
    
    If stable, covergence is slown down to ensure stability.
    """

    err = None
    """
    Maximum estimated error
    """

    geometry = None
    """
    Geometry provided to the solver
    """

    def get_capacitance(self):
        """
        get_capacitance(self)
        
            Get the structure capacitance.
            
            Return:
                Total capacitance [pF].
            
            Note:
                This method can only be used it there are exactly two boundary conditions
                specifying the voltage. Otherwise use :meth:`get_electrostatic_energy` to
                obtain the stored energy :math:`W` and compute the capacitance as:
                :math:`C = 2 \, W / U^2`, where :math:`U` is the applied voltage.
        """

    def get_electrostatic_energy(self):
        """
        get_electrostatic_energy(self)
        
            Get the energy stored in the electrostatic field in the analyzed structure.
            
            Return:
                Total electrostatic energy [J].
        """

    def get_total_current(self, nact=0):
        """
        get_total_current(self, nact=0)
        
            Get total current flowing through active region [mA]
        """

    def get_total_heat(self):
        """
        get_total_heat(self)
        
            Get the total heat produced by the current flowing in the structure.
            
            Return:
                Total produced heat [mW].
        """

    id = None
    """
    Id of the solver object. (read only)
    
    Example:
        >>> mysolver.id
        mysolver:category.type
    """

    inTemperature = None
    """
    Receiver of the temperature required for computations [K].
    
    
    You will find usage details in the documentation of the receiver class
    :class:`~plask.flow.TemperatureReceiverCyl`.
    
    Example:
       Connect the reveiver to a provider from some other solver:
    
       >>> solver.inTemperature = other_solver.outTemperature
    
    See also:
    
       Receciver class: :class:`plask.flow.TemperatureReceiverCyl`
    
       Provider class: :class:`plask.flow.TemperatureProviderCyl`
    
       Data filter: :class:`plask.filter.TemperatureFilterCyl`
    """

    include_empty = None
    """
    Should empty regions (e.g. air) be included into computation domain?
    """

    def initialize(self):
        """
        initialize(self)
        
            Initialize solver.
            
            This method manually initialized the solver and sets :attr:`initialized` to
            *True*. Normally calling it is not necessary, as each solver automatically
            initializes itself when needed.
            
            Returns:
                bool: solver :attr:`initialized` state prior to this method call.
        """

    initialized = None
    """
    True if the solver has been initialized. (read only)
    
    Solvers usually get initialized at the beginning of the computations.
    You can clean the initialization state and free the memory by calling
    the :meth:`invalidate` method.
    """

    def invalidate(self):
        """
        invalidate(self)
        
            Set the solver back to uninitialized state.
            
            This method frees the memory allocated by the solver and sets
            :attr:`initialized` to *False*.
        """

    itererr = None
    """
    Allowed residual iteration for iterative method
    """

    iterlim = None
    """
    Maximum number of iterations for iterative method
    """

    js = None
    """
    Reverse bias current density [A/m\ :sup:`2`\ ].
    
    In case, there is more than one junction you may set $j_s$ parameter for any
    of them by using ``js#`` property, where # is the junction number (specified
    by a role ``junction#`` or ``active#``).
    
    ``js`` is an alias for ``js0``.
    """

    def load_xpl(self, xpl, manager):
        """
        load_xpl(self, xpl, manager)
        
            Load configuration from XPL reader.
            
            This method should be overriden in custom Python solvers.
            
            Example:
              .. code-block:: python
            
                 def load_xpl(self, xpl, manager):
                     for tag in xpl:
                         if tag == 'config':
                             self.a = tag['a']
                             self.b = tag.get('b', 0)
                             if 'c' in tag:
                                 self.c = tag['c']
                         if tag == 'combined':
                             for subtag in tag:
                                 if subtag == 'withtext':
                                     self.data = subtag.attrs
                                     # Text must be read last
                                     self.text = subtag.text
                         elif tag == 'geometry':
                             self.geometry = tag.getitem(manager.geo, 'ref')
        """

    logfreq = None
    """
    Frequency of iteration progress reporting
    """

    maxerr = None
    """
    Limit for the potential updates
    """

    mesh = None
    """
    Mesh provided to the solver
    """

    ncond = None
    """
    Conductivity of the n-contact
    """

    outConductivity = None
    """
    Provider of the computed electrical conductivity [S/m].
    
    
    outConductivity(mesh, interpolation='default')
    
    :param mesh mesh: Target mesh to get the field at.
    :param str interpolation: Requested interpolation method.
    
    :return: Data with the electrical conductivity on the specified mesh **[S/m]**.
    
    Example:
       Connect the provider to a receiver in some other solver:
    
       >>> other_solver.inConductivity = solver.outConductivity
    
       Obtain the provided field:
    
       >>> solver.outConductivity(mesh)
       <plask.Data at 0x1234567>
    
    See also:
    
       Provider class: :class:`plask.flow.ConductivityProviderCyl`
    
       Receciver class: :class:`plask.flow.ConductivityReceiverCyl`
    """

    outCurrentDensity = None
    """
    Provider of the computed current density [kA/cm²].
    
    
    outCurrentDensity(mesh, interpolation='default')
    
    :param mesh mesh: Target mesh to get the field at.
    :param str interpolation: Requested interpolation method.
    
    :return: Data with the current density on the specified mesh **[kA/cm²]**.
    
    Example:
       Connect the provider to a receiver in some other solver:
    
       >>> other_solver.inCurrentDensity = solver.outCurrentDensity
    
       Obtain the provided field:
    
       >>> solver.outCurrentDensity(mesh)
       <plask.Data at 0x1234567>
    
    See also:
    
       Provider class: :class:`plask.flow.CurrentDensityProviderCyl`
    
       Receciver class: :class:`plask.flow.CurrentDensityReceiverCyl`
    """

    outHeat = None
    """
    Provider of the computed heat sources density [W/m³].
    
    
    outHeat(mesh, interpolation='default')
    
    :param mesh mesh: Target mesh to get the field at.
    :param str interpolation: Requested interpolation method.
    
    :return: Data with the heat sources density on the specified mesh **[W/m³]**.
    
    Example:
       Connect the provider to a receiver in some other solver:
    
       >>> other_solver.inHeat = solver.outHeat
    
       Obtain the provided field:
    
       >>> solver.outHeat(mesh)
       <plask.Data at 0x1234567>
    
    See also:
    
       Provider class: :class:`plask.flow.HeatProviderCyl`
    
       Receciver class: :class:`plask.flow.HeatReceiverCyl`
    """

    outPotential = None
    """
    Not available in this solver. Use :attr:`outVoltage` instead.
    """

    outVoltage = None
    """
    Provider of the computed voltage [V].
    
    
    outVoltage(mesh, interpolation='default')
    
    :param mesh mesh: Target mesh to get the field at.
    :param str interpolation: Requested interpolation method.
    
    :return: Data with the voltage on the specified mesh **[V]**.
    
    Example:
       Connect the provider to a receiver in some other solver:
    
       >>> other_solver.inVoltage = solver.outVoltage
    
       Obtain the provided field:
    
       >>> solver.outVoltage(mesh)
       <plask.Data at 0x1234567>
    
    See also:
    
       Provider class: :class:`plask.flow.VoltageProviderCyl`
    
       Receciver class: :class:`plask.flow.VoltageReceiverCyl`
    """

    pcond = None
    """
    Conductivity of the p-contact
    """

    pnjcond = None
    """
    Default effective conductivity of the active region.
    
    Effective junction conductivity will be computed starting from this value.
    Note that the actual junction conductivity after convergence can be obtained
    with :attr:`outConductivity`.
    """

    start_cond = None
    """
    Default effective conductivity of the active region.
    
    Effective junction conductivity will be computed starting from this value.
    Note that the actual junction conductivity after convergence can be obtained
    with :attr:`outConductivity`.
    """

    voltage_boundary = None
    """
    Boundary conditions of the first kind (constant potential) 
    
    This field holds a list of boundary conditions for the solver. You may access
    and alter is elements a normal Python list. Each element is a special class
    that has two attributes:
    
    ============= ==================================================================
    :attr:`place` Boundary condition location (:class:`plask.mesh.RectangularBase2D.Boundary`).
    :attr:`value` Boundary condition value.
    ============= ==================================================================
    
    When you add new boundary condition, you may use two-argument ``append``, or
    ``prepend`` methods, or three-argument ``insert`` method, where you separately
    specify the place and the value. See the below example for clarification.
    
    Example:
        >>> solver.voltage_boundary.clear()
        >>> solver.voltage_boundary.append(solver.mesh.Bottom(), some_value)
        >>> solver.voltage_boundary[0].value = different_value
        >>> solver.voltage_boundary.insert(0, solver.mesh.Top(), new_value)
        >>> solver.voltage_boundary[1].value == different_value
        True
    """

